#include "WPILib.h"
#include "EncoderPID.h"
#include "Math.h"
#include "IOMAP.h"
#include "Hanger.h"
#include "CANJaguar/CANJaguar.h"

//P, I, and D components
#define PC 0.12
#define IC 0.01
#define DC 0.0

//The setpoint the PID loop strives for
#define Setpoint 5

// ASSUMES NO O
// FullADCRange = ((Up - Down) * 1024)
// 1 Rev / 0.5V == 1024 ADC vals / 5V ---> 1 Rev = 1024 * 0.5V / 5V =  102.5 ADC vals
// 4-5 Revs total up -> down
// We want ~ 1 rev per second roughly
// Sample rate = 0.05 sec
// Setpoint....    1 Rev / sec == x / 0.05 sec ->  
//                 x == 0.05 Rev == (0.05 * 102.5) ADC vals 
//                 x ==  5.12 ADC vals  
///  .2  sec samp -> x = 20.5 



	Hanger::Hanger() {
		Raiser = new CANJaguar(HANGER_WINCH);
		Swinger = new CANJaguar(HANGER_TILT);
		Lock = new Servo(HANGER_LOCK);
		
		Potentiometer = new AnalogDifference(HANGER_POT);
		HangerPID = new PIDController(PC, IC, DC, Potentiometer, Raiser);
		
		HangerBroken = false;
		LockEnabled = false;
		
		LastDirection = 0;
		
		/*
		 * HANGER UP/DOWN CONSTANTS HERE
		 * This sets the POT values that represent "full up" and "full down"
		 * The differance should be somewhere between 2.5 and 3 (ish)
		 * (0-5 volts for the full 10 turns, we use 5-6 for our hanger)
		 */
		
		HangerDownSetpoint = 2.5;
		HangerUpSetpoint = .43;
	}
	
	Hanger::~Hanger() {
	    delete HangerPID;
	    delete Potentiometer;
	    delete Lock;
	    delete Swinger;
	    delete Raiser;
		//invoke DrOctagonapus("Bwaaa");
	}
	
	void Hanger::DebugHanger(){
		printf("Pot: %f , UD: %f , LR %f , Srv %f\n "
			   "PID %f , Speed %f , LastPOT %f \n", 
			   Potentiometer->GetVoltage(), 
			   Raiser->Get(), Swinger->Get(),
			   Lock->Get(),
			   HangerPID->Get(),  
			   Potentiometer->getLastSpeed(),
			   Potentiometer->getLastReading());
	}
	
	float Hanger::DebugPot(){
		return Potentiometer->GetVoltage();
	}
	float Hanger::DebugRaiser(){
		return Raiser->Get();
	}
	float Hanger::DebugPID(){
		return HangerPID->Get();
	}
	float Hanger::DebugLock(){
		return Lock->Get();
	}
	double Hanger::DebugSpeed(){
		return Potentiometer->getLastSpeed();
	}
	
	void Hanger::ProcessHanger(){
		if(!HangerBroken && (((Potentiometer->GetVoltage() >= HangerDownSetpoint) && LastDirection == -1) || ((Potentiometer->GetVoltage() <= HangerUpSetpoint) && LastDirection == 1))){
			HangerPID->Reset();
			Raiser->Set(0);
			//printf("Pot voltage: %f, HangerDownSetpoint: %f, HangerBroken: %d, Down Limit: %d\n", Potentiometer->GetVoltage(), HangerDownSetpoint, (int)HangerBroken);			//3.23 at the bottom
		}
	}
	
	//Moves the hange motor up, down, or not at all
	//Pass a 1 for up, a 0 for stop, and a -1 for down
	//Automatically stops the hanger if it has reached it's extent.
	void Hanger::HangerUpDown(int direction){
		LastDirection = direction;
		if(HangerBroken){
			if((direction == 1) && (!LockEnabled)/* && (Potentiometer->GetVoltage() >= HangerUpSetpoint)*/){
				Raiser->Set(.3);
			}
			else{
				if((direction == -1)  /*&& (Potentiometer->GetVoltage() <= HangerDownSetpoint)*/){
					Raiser->Set(-.3);
				}
				else{
					Raiser->Set(0);
				}
			}
		}
		else{
			if((direction == 1) && (!LockEnabled) && (Potentiometer->GetVoltage() >= HangerUpSetpoint)){
				HangerPID->SetSetpoint(Setpoint);
				HangerPID->SetOutputRange(0.0 , 0.4);
				HangerPID->Enable();
			}
			else{
				if((direction == -1)  && (Potentiometer->GetVoltage() <= HangerDownSetpoint)){
					HangerPID->SetSetpoint(-Setpoint);
					HangerPID->SetOutputRange(-1.0 , 0.0);
					HangerPID->Enable();
				}
				else{
					HangerPID->Disable();
				}
			}
		}
	}
	
	void Hanger::HangerLeftRight(int direction){
		Swinger->Set(direction * LeftRightSpeed);
	}
	
	void Hanger::HangerLock(bool lock){
		if(lock){
			Lock->Set(ServoExtent);
			LockEnabled = true;
		}
		else{
			Lock ->Set(0);
			LockEnabled = false;
		}
	}

	void Hanger::SetHangerSensorBroken(bool value){
		HangerBroken = value;
	}
	
	bool Hanger::IsBroken(){
		return HangerBroken;
	}
	void Hanger::Reset(){
		HangerPID->Reset();
		HangerPID->Disable();
		Raiser->Set(0);
	}

