#include "ThunderCam.h"

// Set to 1 to disable use of camera and improve debugging capabilities
// Set to 0 to enable use of camera at the expense of being able to properly
//   reload the program repeatedly on the robot.
#define DISABLE_CAMERA 1


ThunderCam::ThunderCam(DashboardDataSender * dash)
{
#if DISABLE_CAMERA
    camera = NULL;
#else
    camera = ThunderAxisCamera::GetInstance();
#endif
	camVertical = new Servo(CAM_VERT_SERVO);
	dds = dash;
	searchingVisionTarget = true; // default: vision target
	foundTarget = false;
	// initialize null best/alternate targets to start
	bestTargetSoFar.createNull();
	alternateTarget.createNull();
	notSeenCounter = 0;
	timesSeen = 0;
	altTimesSeen = 0;
}
ThunderCam::~ThunderCam() { 
	 // clean up video server so it doesn't bring down the robot
	//ThunderAxisCamera::StopVideoServer();
#if !DISABLE_CAMERA
    ThunderAxisCamera::DeleteInstance();
#endif
	delete camVertical;
}

void ThunderCam::setTargetType(bool isVisionTarget) { 
	searchingVisionTarget = isVisionTarget;
}

bool ThunderCam::searchingForVisionTarget() {
	return searchingVisionTarget;
}

	// tells thundercam to process the targets in the image
void ThunderCam::Process() {
#if !DISABLE_CAMERA
    if (searchingVisionTarget == true) {
		//searching for vision targets...
		
		// if there's a fresh and we're at the previous target heading then
		// get a camera image and process it
		if (camera->IsFreshImage() /*&& turnController.OnTarget() */) {
			// get the gyro heading that goes with this image
			// how to do this?
		
			// get the camera image
			ColorImage *image = camera->GetImage();
			// find FRC targets in the image
			vector<Target> targets = Target::FindCircularTargets(image);
			delete image;
			if (targets.size() == 0 || targets[0].m_score < MINIMUM_SCORE)
			{
				// no targets found. Make sure the first one in the list is 0,0
				// since the dashboard program annotates the first target in green
				// and the others in magenta. With no qualified targets, they'll all
				// be magenta.
				// if no bestTargetSoFar yet, will be null
				if (targets.size() == 0)
					targets.push_back(bestTargetSoFar);
				else	
					targets.insert(targets.begin(), bestTargetSoFar);
//				dds->sendVisionData(0.0, gyro->GetAngle(), 0.0, 0.0, targets);
				dds->sendVisionData(0.0, 0.0, 0.0, 0.0, targets);
/*				if (targets.size() == 0)
					printf("No target found\n\n");
				else
					printf("No valid targets found, best score: %f\n", targets[0].m_score);
				*/
			}
			else {
				// We have some targets.
			
				// search for best target so far among current targets
				bool found = false;
				bool altfound = false;
				for (unsigned i = 0; i <= targets.size(); i++) {
					// have we seen the best target?
					if (bestTargetSoFar.CompareWith(targets[i]) == true)
						found = true;
					// or the alternate?
					if (alternateTarget.CompareWith(targets[i]) == true)
						altfound = true;
//					targets[i].Print();
//					printf("found: %d  altfound: %d", found, altfound);
				}
				if (!found) {
//					printf("best not found\n");
					// best target hasn't been seen again, give it a five
					// frame grace period to be found again
					notSeenCounter++;
					if (notSeenCounter = 5) {
						// grace period is up, clear timesSeen and reset counter
						timesSeen = 0;
						notSeenCounter = 0;
					}
					// only do this if bestTargetSoFar isn't null, otherwise
					// we'd be clearing the alt target every time (and if altfound
					// were true, both bestTargetSoFar and alternateTarget would
					// be cleared)
					if (timesSeen == 0 && bestTargetSoFar.m_majorRadius != 0.0) {
//						printf("alt = best, best cleared\n");
						// we haven't seen it enough, change it to the alternate
						// and clear best
						alternateTarget = bestTargetSoFar;
						bestTargetSoFar.createNull();
						// start search over
					}	
				} else {
//					printf("best found\n");
					// we've found the target again (or nearly the same one)
					// good sign that this target is correct, increment times seen
					timesSeen++;
					notSeenCounter = 0;
				}
				// same with alternate target
				if (!altfound) {
//					printf("alt not found\n");
					if (altTimesSeen > 0) altTimesSeen--;
				} else {
//					printf("alt found\n");
					altTimesSeen++;
					// have we seen it a lot more than the best one?
					if (altTimesSeen > timesSeen) {
//						printf("switch alt and best\n");
						// switch it with the best target
						Target temptarget = alternateTarget; 
						alternateTarget = bestTargetSoFar;
						bestTargetSoFar = temptarget;
					}
				}
				// best target in vector will be most likely to be the vision
				// target because it will have highest score;
				// is it different from previous best?
				if (bestTargetSoFar.CompareWith(targets[0]) == false) {
					// is it different from the alternate?
					if (alternateTarget.CompareWith(targets[0]) == false) {
						// is it better or at least about equal?
						if (targets[0].m_score > (alternateTarget.m_score * 0.9)) {
//							printf("new alt\n");
							alternateTarget = targets[0]; // yes- new alternate
						}
					}
				}
//				targets[0].Print(); // for debug
				// now replace best target in vector with the best target
				// so far so it doesn't jump around
				targets[0] = bestTargetSoFar;
			
				// send dashboard data for target tracking
				dds->sendVisionData(0.0, 0.0, 0.0, targets[0].m_xPos / targets[0].m_xMax, targets);
//				bestTargetSoFar.Print();
//				alternateTarget.Print();
			} //(targets.size() == 0 || targets[0].m_score < MINIMUM_SCORE)
		foundTarget = false;
		// If target is good enough or we've seen it often enough (tweak the value?),
		// we probably have found the vision target
		if (bestTargetSoFar.m_bothFound == true || timesSeen > 5) foundTarget = true;
		} // camera->IsFreshImage()
	} else { // !(searchingVisionTarget == true)
		// searching for balls... will probably never be implemented
	}
#endif
}
	
bool ThunderCam::targetFound() {
	return foundTarget;
}

float ThunderCam::getAngleToTarget() {
//	printf("%f\n",(double)horizontalAngle);
	if (!foundTarget) return 9001;
	//use best target and get its horizontal distance from image center
	// -- value is given as fraction of image height, and is 2x as large as real
	//    x distance
	double x = bestTargetSoFar.m_xPos * bestTargetSoFar.m_imageHeight / 2.0;
//	printf("X: %f", x);
	// horizontal distance from camera to target
	double distance = getDistanceToTarget();
	// convert from image size to "real" size -- see documentation
	x = (x * distance) / FOCAL_LENGTH;
	// angle is arc sine of "real" x and distance to target
	// convert from radians to degrees
	double angle = asin(x / distance) / 3.1415926535 * 180.0;
//	printf("Xreal: %f,  Dh: %f, theta: %f\n", x, distance, angle);
	return (float)angle;
}

float ThunderCam::getDistanceToTarget() {
	if (!foundTarget) return -9001;
	//use best target and get its height -- is 2 times the radius
	// which is given as a fraction of the image height
	double height = 2.0 * bestTargetSoFar.m_majorRadius * bestTargetSoFar.m_imageHeight;
	// if height is shorter than width -- rare occurrence --
	// then m_rotation will be either around 0 or 180
	if ((fabs(bestTargetSoFar.m_rotation) < 45.0) || 
			(fabs(bestTargetSoFar.m_rotation-180.0) < 45.0)) {
		height = 2.0 * bestTargetSoFar.m_minorRadius * bestTargetSoFar.m_imageHeight;
	}
	// see documentation for complex math:
	// intermediate value
	double x = (2.0 * height * DELTA_HEIGHT) / (FOCAL_LENGTH * TARGET_SIZE);
	// vertical angle camera to target
	double alpha = (asin(x)) / 2.0;
	// horizontal distance to target
	double distance = (DELTA_HEIGHT / tan(alpha));
//	printf("X: %f   Alpha: %f   Dh:	%f\n", x, alpha, distance);
	return (float)distance;
}

void ThunderCam::setServoValue(float value) {
	// using servo values directly, since no one will really
	// be interested in seeing the actual camera angle(?)
	if ((value <= CAM_MAX_SERVO_VALUE) &&
			(value >= CAM_MIN_SERVO_VALUE)) {
		// make sure servo value is in acceptable range so
		// camera stays in range horizontal to 30 degrees down
		camVertical->Set(value);
	} else {
		printf("Bad servo value %f", value);
	}
}

float ThunderCam::getServoValue() {
	return camVertical->Get();
}


